import unittest, time
from copy import copy

from kcapi.rest.crud import KeycloakCRUD
from kcapi.rest.targets import Targets
from kcapi.rest.url import RestURL
from .testbed import KcBaseTestCase
import json
import os

def load_sample(fname):
    f = open(fname)
    file1 = json.loads(f.read())
    f.close()
    return file1

def exist(that, res, name):
    ret  = res.findFirstByKV('username', name)
    that.assertEqual(name, ret['firstName'], 'We expect a role ['+name+'] to be created')
    return ret['id']



def test_complete_CRUD(that, users):
        ## POST
        state = users.create(that.USER_DATA)
        that.assertTrue(state, 'fail while posting')

        ret  = users.findFirstByKV('username', 'pepe')
        that.assertEqual('pepe', ret['firstName'], 'We expect a user with pepe as username to be created')

        ## TEST existByKV True
        existByKV_true_state = users.existByKV("username",that.USER_DATA["username"])
        that.assertTrue(existByKV_true_state,msg="[TEST existByKV True] We expect True if user does exist")

        ## UPDATE
        state = users.update(ret['id'], {'firstName': 'pedro'})
        that.assertTrue(state, 'fail while updating')

        ret  = users.findFirstByKV('firstName', 'pedro')
        that.assertTrue(ret != False, 'Something wrong updating the resource.')
        that.assertEqual('pedro', ret['firstName'], 'We expect a user with pepe as username to be created')

        ## GET
        usr = users.get(ret['id']).resp().json()
        that.assertEqual(ret['id'], usr['id'])

        _all = users.all()
        that.assertEqual(len(_all), 4, 'All users amount to one')


        ## DELETE
        remove_state = users.remove(ret['id']).isOk()
        that.assertTrue(remove_state)
        removed = users.findFirstByKV('firstName', 'pedro')

        that.assertFalse(removed)

        ## TEST existByKV False
        existByKV_false_state = users.existByKV("username","user_existByKV")
        that.assertFalse(existByKV_false_state,msg="[TEST existByKV False] We expect False if user does not exist")


class SlowFile:
    def __init__(self, content, delays=[]):
        self.content = bytes(content, 'utf-8')
        self.delays = delays
        self.ii = -1

    def __iter__(self):
        # self.ii = -1
        return self

    def __next__(self):
        self.ii += 1
        if self.ii >= len(self.content):
            raise StopIteration
        if self.ii < len(self.delays):
            time.sleep(self.delays[self.ii])
        return self.content[self.ii: self.ii+1]


class Testing_slow_crud(KcBaseTestCase):

    def testing_crud_API(self):
        token = self.testbed.token

        users = KeycloakCRUD()
        users.targets = Targets.makeWithURL(str(self.USER_ENDPOINT))
        users.token = token

        test_complete_CRUD(self, users)

    @unittest.skipIf(not int(os.environ.get("KC_RUN_SLOW_TEST", "0")), "Test is run only if KC_RUN_SLOW_TEST is set")
    def test_slow_CRUD(self):
        """
        We test if long-lasting requests can complete.
        We monkey patch requests.put to achieve long upload time.
        The access_token expires in 60 sec.
        Assumption is that access_token is verified at beginning of request only.
        Verification: run
        KC_RUN_SLOW_TEST=1 python -m unittest test.test_crud.Testing_User_API.test_slow_CRUD
        and check TCP traffic with tcpdump.
        """
        token = self.testbed.token
        users = KeycloakCRUD()
        users.targets = Targets.makeWithURL(str(self.USER_ENDPOINT))
        users.token = token

        ## IS PRESENT
        user_present = users.findFirstByKV('firstName', 'pepe')
        self.assertFalse(user_present)

        ## POST
        # state = users.create(self.USER_DATA)
        # inline code for users.create()
        def users_create(obj, payload):
            import requests
            from kcapi.rest.resp import ResponseHandler
            url = obj.targets.url('create')
            data_file = SlowFile(json.dumps(payload), [10]*10)
            with KcSession() as session:
                ret = session.post(url, data=data_file, headers=obj.headers())
            return ResponseHandler(url, method='Post', payload=payload).handleResponse(ret)

        state = users_create(users, self.USER_DATA)
        self.assertTrue(state, 'fail while posting')

        ret = users.findFirstByKV('username', 'pepe')
        self.assertEqual('pepe', ret['firstName'], 'We expect a user with pepe as username to be created')

        ## DELETE
        remove_state = users.remove(ret['id']).isOk()
        self.assertTrue(remove_state)
        ## IS PRESENT
        user_present = users.findFirstByKV('firstName', 'pepe')
        self.assertFalse(user_present)

    def setUp(self):
        super().setUp()

        self.USER_ENDPOINT = RestURL(self.testbed.ENDPOINT)
        self.USER_ENDPOINT.addResources(['auth', 'admin', 'realms', self.REALM, 'users'])

        self.USER_DATA = {"enabled":True,"attributes":{},"username":"pepe","emailVerified":"", "firstName": 'pepe'}


class TestKeycloakCRUD_part2(KcBaseTestCase):
    def setUp(self):
        super().setUp(create_all=False)
        self.REALM = self.testbed.REALM
        self.testbed.createRealms()
        self.testbed.createUsers()

    def test_get_one_ok(self):
        users_api = self.testbed.kc.build("users", self.testbed.REALM)
        user_id = users_api.findFirstByKV("username", "batman")["id"]

        user = users_api.get_one(user_id)

        self.assertIsInstance(user, dict)
        # "assertDictContainsSubset(a, b)" == "a == a|b"
        self.assertEqual(user, user | {
            "username": "batman",
            "lastName": "Wayne",
            "enabled": True,
        })

    def test_get_one_invalid_id(self):
        users_api = self.testbed.kc.build('users', self.testbed.REALM)
        self.assertRaises(Exception, users_api.get_one, "no-such-user-uuid")

    def test_update_rmw_ok(self):
        users_api = self.testbed.kc.build("users", self.testbed.REALM)
        user_old = users_api.findFirstByKV("username", "batman")
        user_id = user_old["id"]

        # data1 = {"username": "batman-new"}  # username cannot be updated, but API return 204
        data1 = {"lastName": "Wayne-new"}
        data2 = {"email": "bw@example.com"}
        data3 = {"enabled": False}

        ret = users_api.update_rmw(user_id, {})
        ret = users_api.update_rmw(user_id, data1)
        ret = users_api.update_rmw(user_id, data2)
        ret = users_api.update_rmw(user_id, data3)

        user_new = users_api.get_one(user_id)
        # "assertDictContainsSubset(a, b)" == "a == a|b"
        self.assertEqual(user_new, user_old | {
            "username": "batman",
            "lastName": "Wayne-new",
            "email": "bw@example.com",
            "enabled": False,
        })

    def test_update_rmw_invalid_id(self):
        users_api = self.testbed.kc.build('users', self.testbed.REALM)
        self.assertRaises(Exception, users_api.update_rmw, "no-such-user-uuid", {})


if __name__ == '__main__':
    unittest.main()

# setUp vs setUpClass



import os
import requests


class KcSession(requests.Session):
    """
    Small wrapper around requests.Session.
    KEYCLOAK_API_CA_BUNDLE environ variable can be used to:
    - disable certificate verification (KEYCLOAK_API_CA_BUNDLE=""), or,
    - use alternative CA bundle (KEYCLOAK_API_CA_BUNDLE="/path/to/file-or-dir").
    """
    def __init__(self):
        super().__init__()
        # similar to REQUESTS_CA_BUNDLE
        if "KEYCLOAK_API_CA_BUNDLE" in os.environ:
            if os.environ["KEYCLOAK_API_CA_BUNDLE"]:
                self.verify = os.environ["KEYCLOAK_API_CA_BUNDLE"]
            else:
                self.verify = False



from copy import deepcopy

from .crud import KeycloakCRUD


class Role():
    def __init__(self, kc, role):
        self.value = role

        # Note: Composites() kc param must be top-level API URL
        # deepcopy is needed, otherwise we end with corrupted parent client_api (the shallow
        # copied URLs would get patched to be suitable for client role)
        kc_top_level = deepcopy(kc)
        for rest_method in kc_top_level.targets.targets:
            custom_url = kc_top_level.targets.targets[rest_method].copy()
            assert custom_url.resources[-1] == 'clients'
            custom_url.removeLast()
            kc_top_level.targets.targets[rest_method] = custom_url

        self.kc = kc_top_level

    def composite(self):
        return Composites(self.kc, self.value['id'])


class ClientRoleCRUD(KeycloakCRUD):
    """
    For RH SSO 7.4 (KC 9.0), role with attributes cannot be created.
    We need to first create role, then update it to add attributes.

    This will not be needed any more in RH SSO 7.5.
    """
    def create(self, payload):
        ret = super().create(payload)
        if "attributes" in payload:
            role = self.findFirstByKV("name", payload["name"])
            self.update(role["id"], payload).isOk()
        return ret


class Composites:
    def __init__(self, kc, roleID):
        # Note: kc must be top-level API URL
        self.kc = kc
        self.id = roleID

        # It just adds few extra resource to the url: /id/composites
        roles_by_id_api = KeycloakCRUD.get_child(kc, None, "roles-by-id")
        self.post = lambda payload: KeycloakCRUD.get_child(roles_by_id_api, self.id, 'composites').create(payload)
        self.remove = lambda payload: KeycloakCRUD.get_child(roles_by_id_api, self.id, 'composites').remove(_id=None, payload=payload)
        # It just adds few extra resource to the url: /id/composites/realm
        # Hm. This assumes composite client role will contain only realm roles (not other client roles).
        # Is this always true, does this cover all usecases?
        self.get = lambda: KeycloakCRUD.get_child(roles_by_id_api, self.id, 'composites/realm').findAll()
        self.get_role_by_name = lambda name='': KeycloakCRUD.get_child(kc, None, 'roles').findFirstByKV(key='name', value=name)

    def link(self, roleName):
        role = self.get_role_by_name(roleName)

        if not role:
            raise("Error the role "+roleName+" not found!")

        return self.post([role])

    def unlink(self, roleName):
        role = self.get_role_by_name(roleName)

        if not role:
            raise ("Error the role " + roleName + " not found!")

        # Whole role representation needed in DELETE payload
        # See https://www.keycloak.org/docs-api/20.0.1/rest-api/index.html#_roles_by_id_resource
        return self.remove([role])

    def findAll(self):
        return self.get()


def new_child(kc, query, child_resource):
    client_id = kc.findFirst(query)['id']
    return KeycloakCRUD.get_child(kc, client_id, child_resource)


class Clients(KeycloakCRUD):

    def secrets(self, client_query):
        obj = super().findFirst(client_query)
        child = KeycloakCRUD.get_child(self, obj['id'], 'client-secret')
        return child

    def get_roles(self, client_query):
        # TODO maybe move briefRepresentation=False into the RestURL class?
        # So we always get same object returned.
        roles = self.roles(client_query).findAll(dict(briefRepresentation=False)).resp().json()

        assert str(self.targets.targets["read"]).endswith("/clients")
        ret = [Role(self, role) for role in roles]
        assert str(self.targets.targets["read"]).endswith("/clients")

        return ret

    def roles(self, client_query):
        client_id = super().findFirst(client_query)['id']
        client_roles_api = ClientRoleCRUD.get_child(self, client_id, 'roles')
        client_roles_api = self.__hack_rest_roles_remove_and_update_endpoint(client_roles_api)
        return client_roles_api

    def __hack_rest_roles_remove_and_update_endpoint(self, client_roles_api):
        custom_delete = self.targets.targets['delete'].copy()
        custom_delete.replaceResource('clients', 'roles-by-id')
        client_roles_api.targets.targets['delete'] = custom_delete

        custom_update = self.targets.targets['update'].copy()
        custom_update.replaceResource('clients', 'roles-by-id')
        client_roles_api.targets.targets['update'] = custom_update

        return client_roles_api



import unittest
import json
from copy import copy

from .testbed import KcBaseTestCase


def load_sample(fname):
    f = open(fname)
    file1 = json.loads(f.read())
    f.close()
    return file1


class TestClients(KcBaseTestCase):
    def test_client_extra_methods(self):
        clients = self.testbed.getKeycloak().build('clients', self.REALM)
        self.assertIsNotNone(clients.roles, 'The client object should have a roles method.')

    def test_client_secrets(self):
        secret_passphrase = "cf6c3215-2af0-42f7-a9ce-3a371e738417"
        clients = self.testbed.getKeycloak().build('clients', self.REALM)
        state = clients.create({"enabled":True,"attributes":{},"secret":secret_passphrase ,"redirectUris":[],"clientId":"44","protocol":"openid-connect", "publicClient": False}).isOk()
        self.assertTrue(state, "A private client should be created")

        #client_secrets = clients.secrets({'key': 'clientId', 'value':"44"})
        #remote_passphrase = client_secrets.all()

        #client_secrets_creation_state = client_secrets.create({"type":self.testbed.REALM,"client":secret_passphrase}).isOk()
        #self.assertTrue(client_secrets_creation_state, "A secret passphrase should be assigned.")
        #remote_passphrase = client_secrets.all()
        #self.assertTrue(len(remote_passphrase)>0, "We expect to obtain a new passphrase.")

    def test_client_roles(self):
        client_payload = load_sample('./test/payloads/client.json')
        clients_api = self.testbed.getKeycloak().build('clients', self.REALM)

        # initial cleanup
        ret = clients_api.removeFirstByKV('clientId', client_payload['clientId'])
        ret = clients_api.findFirstByKV('clientId', client_payload['clientId'])
        self.assertEqual(ret, [], 'Client should not exist before')
        # create empty client
        svc_state = clients_api.create(client_payload).isOk()
        self.assertTrue(svc_state, 'The service should return a 200.')

        ret = clients_api.findFirstByKV('clientId', client_payload['clientId'])
        self.assertNotEqual(ret, [], 'It should return the posted client')

        # test client role create, with attributes
        client_query = {'key': 'clientId', 'value': client_payload['clientId']}
        client_roles_api = clients_api.roles(client_query)
        role_doc = {
            "name": "new-role",
            "description": "here should go a description.",
            "attributes": {
                "ci-new-role-key0": [
                    "ci-new-role-value0"
                ]
            },
            'clientRole': True,
            'composite': False,
        }
        svc_roles_state = client_roles_api.create(role_doc).isOk()
        self.assertTrue(svc_roles_state, 'The client_roles service should return a 200.')

        client_roles = clients_api.get_roles(client_query)
        self.assertEqual(1, len(client_roles))
        new_role = client_roles[0]
        new_role_min = copy(new_role.value)
        new_role_min.pop("id")
        new_role_min.pop("containerId")
        self.assertEqual(role_doc, new_role_min)

        # create realm role, it will be added as composite (sub-role) to client role
        role = {"name": "x_black_magic_x"}
        roles = self.testbed.getKeycloak().build("roles", self.REALM)
        state = roles.create(role).isOk()
        self.assertTrue(state, "A role should be created")

        # test adding composite/sub-role
        state = new_role.composite().link(role['name']).isOk()

        self.assertTrue(state, "A composite role should be added to the current client role.")

        composites_roles = new_role.composite().findAll().resp().json()
        self.assertEqual(composites_roles[0]['name'], 'x_black_magic_x', "We should have a composite role called: x_black_magic_x")

        # test removing composite/sub-role
        state_delete = new_role.composite().unlink(role['name']).isOk()
        self.assertTrue(state, "A composite role should be deleted from the current client role.")

        empty_composites_role = new_role.composite().findAll().resp().json()
        self.assertEqual(empty_composites_role, [],
                         "We should have a empty roles")

    def test_client_roles_update(self):
        client_payload = load_sample('./test/payloads/client.json')
        clients_api = self.testbed.getKeycloak().build('clients', self.REALM)

        # initial cleanup
        ret = clients_api.removeFirstByKV('clientId', client_payload['clientId'])
        ret = clients_api.findFirstByKV('clientId', client_payload['clientId'])
        self.assertEqual(ret, [], 'Client should not exist before')
        # create empty client
        svc_state = clients_api.create(client_payload).isOk()
        self.assertTrue(svc_state, 'The service should return a 200.')

        ret = clients_api.findFirstByKV('clientId', client_payload['clientId'])
        self.assertNotEqual(ret, [], 'It should return the posted client')

        # create role
        client_query = {'key': 'clientId', 'value': client_payload['clientId']}
        client_roles_api = clients_api.roles(client_query)
        svc_roles_state = client_roles_api.create(
            {"name": "new-role", "description": "here should go a description."}).isOk()
        self.assertTrue(svc_roles_state, 'The client_roles service should return a 200.')
        # ----

        # check initial state
        assert str(clients_api.targets.targets["read"]).endswith("/clients")
        # is ok
        new_role_a = clients_api.get_roles(client_query)[0]
        # URL is corrupted
        assert str(clients_api.targets.targets["read"]).endswith("/clients")

        new_role_a_min = copy(new_role_a.value)
        new_role_a_min.pop("id")
        new_role_a_min.pop("containerId")
        self.assertEqual(
            {
                'attributes': {},
                'clientRole': True,
                'composite': False,
                'description': 'here should go a description.',
                'name': 'new-role'
            },
            new_role_a_min,
        )

        # update role
        # We need to send full payload.
        new_data = copy(new_role_a.value)
        new_data.update({
            "description": "here should go a description. NEW",
            "attributes": {
                "ci-new-role-key0": [
                    "ci-new-role-value0"
                ]
            },
        })
        client_roles_api.update(new_role_a.value["id"], new_data).isOk()

        # check updated state
        new_role_b = clients_api.get_roles(client_query)[0]
        new_role_b_min = copy(new_role_b.value)
        new_role_b_min.pop("id")
        new_role_b_min.pop("containerId")
        self.assertEqual(
            {
                'attributes': {'ci-new-role-key0': ['ci-new-role-value0']},
                'clientRole': True,
                'composite': False,
                'description': 'here should go a description. NEW',
                'name': 'new-role'
            },
            new_role_b_min,
        )

    def test_client_roles_removal(self):
        client_payload = load_sample('./test/payloads/client.json')
        client_payload['clientId'] = 'test_client_roles_removal'
        client_role_name = "deleteme-role"

        clients_api = self.testbed.getKeycloak().build('clients', self.REALM)
        svc_state = clients_api.create(client_payload).isOk()

        client_query = {'key': 'clientId', 'value': client_payload['clientId']}
        client_roles_api = clients_api.roles(client_query)
        svc_roles_state = client_roles_api.create(
            {"name": client_role_name, "description": "A role that need to be deleted."}).isOk()

        ret_client_roles = client_roles_api.findFirstByKV('name', client_role_name)
        self.assertNotEqual(ret_client_roles, [], 'It should return the posted client')

        result = client_roles_api.removeFirstByKV('name', client_role_name)
        self.assertTrue(result, 'The server should return ok.')

        self.assertEqual(client_roles_api.findFirstByKV('name', client_role_name), [], 'It should return the posted client')

    # TODO - likely all tests should use only setUp/tearDown
    #  - each test has clean environment
    #  - VCR than can really work
    # @classmethod
    # def setUpClass(cls):
    #     pass

    # @classmethod
    # def tearDownClass(cls):
    #     pass

    def setUp(self):
        super().setUp(create_all=False)
        self.testbed.createRealms()
        # not needed
        # cls.testbed.createUsers()
        # cls.testbed.createClients()
        self.REALM = self.testbed.REALM

    def tearDown(self):
        self.testbed.goodBye()



from kcapi import OpenID, Keycloak
import os
import json
import unittest
from vcr_unittest import VCRTestCase

import logging
logging.basicConfig(level=logging.DEBUG)

def readFromJSON(filename):
    with open(filename) as json_file:
        return json.load(json_file)

class TestBed:
    def __init__(self, realm = None, username = None, password = None, endpoint = None): 

        self.USER = os.environ['KC_USER']
        self.PASSWORD = os.environ['KC_PASSWORD']
        self.REALM = os.environ['KC_REALM']
        self.ENDPOINT = os.environ['KC_ENDPOINT']

        self.groupName = 'DC'
        self.roleNames = ['level-1', 'level-2', 'level-3'] 

        # TODO this is not intercepted by vcrpy.
        token = OpenID.createAdminClient(self.USER, self.PASSWORD, url=self.ENDPOINT).getToken()
        self.kc = Keycloak(token, self.ENDPOINT)
        self.master_realm = self.kc.admin()
        self.realm = self.REALM 
        self.token = token

    def deleteRealms(self):
        realm = self.realm
        if self.master_realm.existByKV("id", realm):
            self.master_realm.removeFirstByKV("id", realm)

    def createRealms(self):
        realm = self.realm
        self.master_realm.create({"enabled": "true", "id": realm, "realm": realm})

    def createGroups(self):
        group = self.kc.build('groups', self.realm)
        g_creation_state = group.create({"name": self.groupName}).isOk()
        self.createRoles()


    def createRoles(self):
        roles = self.kc.build('roles', self.realm)
        for role in self.roleNames:
            roles.create({"name": role}).isOk() 


    def createClients(self):
        realm = self.realm
        client = {"enabled":True,
                  "attributes":{},
                  "redirectUris":[],
                  "clientId":"dc",
                  "protocol":"openid-connect", 
                  "directAccessGrantsEnabled":True
                  }

        clients = self.kc.build('clients', realm)
        if not clients.create(client).isOk(): 
            raise Exception('Cannot create Client')

    def createUsers(self):
        realm = self.realm
        test_users = [
                {"enabled":'true',"attributes":{},"username":"batman","firstName":"Bruce", "lastName":"Wayne", "emailVerified":""}, 
                {"enabled":'true',"attributes":{},"username":"superman","firstName":"Clark", "lastName":"Kent", "emailVerified":""}, 
                {"enabled":'true',"attributes":{},"username":"aquaman","firstName":"AAA%", "lastName":"Corrupt", "emailVerified":""}
        ]

        users = self.kc.build('users', realm)
        
        for usr in test_users: 
            users.create(usr).isOk()

    def goodBye(self):
        if self.master_realm.exist(self.realm):
            state = self.master_realm.remove(self.realm).ok()
            if not state:
                raise Exception("Cannot delete the realm -> " + self.realm )

    def getKeycloak(self):
        return self.kc
    def getAdminRealm(self):
        return self.master_realm


class KcBaseTestCase(VCRTestCase):
    # see https://github.com/agriffis/vcrpy-unittest/blob/master/vcr_unittest/testcase.py#L19
    # Try self.vcr_enabled=False to disable tests

    def setUp(self, *, create_all=True):
        super().setUp()
        self.testbed = TestBed()
        if create_all:
            self.testbed.createRealms()
            self.testbed.createUsers()
            self.testbed.createClients()
            self.REALM = self.testbed.REALM

    def tearDown(self):
        self.testbed.goodBye()
        super().tearDown()

    def _get_vcr_kwargs(self):
        kwargs = super()._get_vcr_kwargs()
        # https://vcrpy.readthedocs.io/en/latest/usage.html#record-modes
        # "once" or "new_episodes" makes sense - which is better?
        # "new_episodes" is more auto-magic, use it.
        # BUT - authentiaction is in setUp(), is not recorded, so each time a diffrent access token is send,
        # and VCR is not used.
        # We use "none".
        vcrpy_record_mode = os.environ.get("VCRPY_RECORD_MODE", "none")
        kwargs['record_mode'] = vcrpy_record_mode
        return kwargs



import unittest

from .testbed import KcBaseTestCase


class TestServerInfo(KcBaseTestCase):
    def test_serverinfo_api(self):
        kc = self.testbed.getKeycloak()
        serverinfo_api = kc.build_serverinfo()

        serverinfo = serverinfo_api.get(None).verify().resp().json()

        expected_server_version_all = [
            "9.0.3",
            "15.0.2",
        ]

        self.assertIsInstance(serverinfo, dict)
        self.assertIn(serverinfo["systemInfo"]["version"], expected_server_version_all)
        expected_sv = serverinfo["systemInfo"]["version"]  # "15.0.2"
        expected_sv_p1 = '.'.join(expected_sv.split('.')[0:1])  # "15"
        expected_sv_p2 = '.'.join(expected_sv.split('.')[0:2])  # "15.0"
        # expected_sv_p2 = '.'.join(expected_sv.split('.')[0:3])

        self.assertEqual(expected_sv, serverinfo["systemInfo"]["version"])
        self.assertEqual("community", serverinfo["profileInfo"]["name"])
        self.assertEqual(expected_sv, kc.server_info.version)
        self.assertEqual("community", kc.server_info.profile_name)

        self.assertEqual("community " + expected_sv_p2, kc.server_info_compound_profile_version())
        self.assertEqual("community " + expected_sv_p1, kc.server_info_compound_profile_version(1))
        self.assertEqual("community " + expected_sv_p2, kc.server_info_compound_profile_version(2))
        self.assertEqual("community " + expected_sv, kc.server_info_compound_profile_version(3))
        self.assertEqual("community " + expected_sv, kc.server_info_compound_profile_version(10))



from copy import copy


# def create_flow(flow):
#     alias = flow['displayName']
#     pid = None if not 'provider' in flow else flow['providerId']
#     provider = 'registration-page-flow' if not pid else pid
#     flow_type = 'basic-flow' if not pid else 'form-flow'
#
#     # WARN: The value description is not well validated in Keycloak, it can return 500.
#     return {
#         "alias": alias,
#         "type": flow_type,
#         "description": "empty",
#         "provider": provider,
#     }

# replacement for create_flow()
def create_child_flow_data(flow):
    """
    :param flow: What was stored in json file.
    :return: What needs to be POST-ed to server to create a flow.
    Look at kcfetcher inject_data.py for POST/.create() examples.
    """
    assert 'authenticationFlow' in flow and flow['authenticationFlow'] is True
    assert 'topLevel' not in flow

    # https://172.17.0.3:8443/auth/admin/realms/ci0-realm/authentication/flows
    # description is returned, they are flows, and they are top level flows.
    # topLevel==true
    # Payload to create those top-level flows is NOT computed here.
    #
    # https://172.17.0.3:8443/auth/admin/realms/ci0-realm/authentication/flows/ci0-auth-flow-generic/executions
    # They can be child flows or child executions
    # Payload to create child executions is NOT computed here.
    # Here is computed only payload for child flows.
    #
    # API reponse for child flow
    # - there is no alias, but it has displayName.
    # - no topLevel attribute
    # - has level in index attribute
    #
    # So: for child flows, on POST/.create() we send some alias.
    # On GET/.get() alias attribute is not returned is not returned, but value of alias is in displayName.
    assert "alias" not in flow
    alias = flow['displayName']

    # In .json (API GET response) we have "providerId", to server we send (API POST) "provider".
    pid = flow.get('providerId')

    # Child flows are created with "provider"=="registration-page-form"
    # Child executions have providerId stored in json.
    # provider = pid or 'registration-page-flow'  # orig concept
    assert pid != 'registration-page-flow'  # just want to know if 'registration-page-flow' is ever used.
    provider = pid or 'registration-page-form'  # this will work on ci0-auth-flow-generic-exec-3-generic-alias case

    # flow_type = 'basic-flow' if not pid else 'form-flow'  # orig code
    if pid:
        flow_type = 'form-flow'
    else:
        flow_type = 'basic-flow'

    assert "description" not in flow
    # Real description is not even stored by kcfethcer
    # Seems real description is not even returned by API - is not observable, likely we can just ignore it
    # WARN: The value description is not well validated in Keycloak, it can return 500.
    data = {
        "alias": alias,
        "type": flow_type,
        "description": "empty-description",
        "provider": provider,
    }
    return data


def get_executions(execution):
    provider = execution['providerId']
    return {'provider': provider}


def is_auth_flow(body):
    if 'authenticationFlow' in body:
        assert body['authenticationFlow'] is True
    return 'authenticationFlow' in body


def get_node_level(node):
    return node['level']


def consistent(parent, flow):
    published_flows = parent.all()

    for publishedFlow in published_flows:
        resource_id = False

        if 'providerId' in publishedFlow and 'providerId' in flow:
            resource_id = publishedFlow['providerId'] == flow['providerId']
        else:
            resource_id = publishedFlow['displayName'] == flow['displayName']

        level_is_equal = publishedFlow['level'] == flow['level']
        index_is_equal = publishedFlow['index'] == flow['index']

        if resource_id and level_is_equal and index_is_equal:
            publishedFlow['requirement'] = flow['requirement']
            parent.update(None, publishedFlow).isOk()
            return True

    return False


def get_node_identity_pair(root_node):
    key = 'alias' if 'alias' in root_node else 'displayName'
    unique_identifier = root_node[key]

    return [key, unique_identifier]

def get_matching_flow_from_server(flow, stored_flows):
    [_, flow_id] = get_node_identity_pair(flow)

    for stored_flow in stored_flows:
        [_, id] = get_node_identity_pair(stored_flow)
        if flow_id == id:
            return stored_flow

    raise Exception("Trying to updated a built-in flow, with the wrong flow: " + flow + ". \n Make sure that this flow is defined in the built-in flow.")


def merge_flows(flow_from_server = {}, local_flow={}):
    changed = False
    for key in local_flow.keys():
        if key in flow_from_server and flow_from_server[key] != local_flow[key] and key != 'flowId':
            flow_from_server[key] = local_flow[key]
            changed = True

    return [flow_from_server, changed]

class AuthenticationFlowsImporter():
    def __init__(self, authentication_api):
        self.flowAPI = authentication_api

    def update(self, root_node, flows):
        # flows - child flows (from executors.json) belonging to root_node top-level flow.
        # TODO setup configj for flows/executions that were configured.
        assert root_node["topLevel"] is True

        flow_api = self.flowAPI.flows(root_node)
        executions_api = self.flowAPI.executions(root_node)

        if self.flowAPI.is_built_in(root_node):
            stored_flows = flow_api.all()
            #stored_executions = executions_api.all()

            for flow in flows:
                if is_auth_flow(flow):
                    flow_from_server = get_matching_flow_from_server(flow, stored_flows)
                    [updated_flow, has_change] = merge_flows(flow_from_server, flow)

                    if has_change:
                        flow_api.update(None, updated_flow).isOk()
                else:
                    flow_min = copy(flow)
                    flow_min.pop("authenticationConfigData", {})
                    executions_api.update(None, flow_min).isOk()
        else:
            # Create top-level flow.
            # TODO - remove/create only if change is needed.
            self.remove(root_node)
            self.flowAPI.create(root_node).isOk()
            # Now create child flows.
            self.publish(root_node, flows)

    def remove(self, root_node):
        [key, unique_identifier_value] = get_node_identity_pair(root_node)
        self.flowAPI.removeFirstByKV(key, unique_identifier_value)

    def publish(self, root_node, flows):
        if not isinstance(flows, list):
            raise Exception("Bad Parameters for Authentication Flow: auth_flow parameter should be an array.")

        root_node = root_node
        nodes = {0: root_node}

        root_flow = self.flowAPI.executions(root_node)

        for flow in flows:
            current_level = get_node_level(flow)
            parent = nodes[current_level]

            if is_auth_flow(flow):
                authentication_flow = create_child_flow_data(flow)
                nodes[current_level + 1] = authentication_flow
                self.flowAPI.flows(parent).create(authentication_flow).isOk()
            else:
                execution = get_executions(flow)
                self.flowAPI.executions(parent).create(execution).isOk()

            if not consistent(root_flow, flow):
                raise Exception(
                    'There is an inconsistency problem: Changes are not taking place in the server, latency problems ?.')


"""
debugging
what is created by kcfetcher inject_data.py

1
execution
authenticationFlow missing
"providerId": "direct-grant-validate-username"

2
execution
authenticationFlow missing
"displayName": "Conditional OTP Form",
"providerId": "auth-conditional-otp-form",
"alias": "ci0-auth-flow-generic-exec-20-alias",
    alias was set when config was added

3
flow, "type": "basic-flow",
"authenticationFlow": true,
"displayName": "ci0-auth-flow-generic-exec-3-generic-alias"
    POST - alias was stored into displayName
providerId missing

4
flow, "type": "basic-flow", child of n3
"authenticationFlow": true,
"displayName": "ci0-auth-flow-generic-exec-3-1-flow-alias"
providerId missing

5
flow,  "type": "form-flow",
"authenticationFlow": true,
"displayName": "ci0-auth-flow-generic-exec-4-flow-alias"
"providerId": "registration-page-form",

6
execution, child od 4
authenticationFlow missing
"alias": "ci0-auth-flow-generic-exec-6-alias"
    alias was set when config was added
"displayName": "Recaptcha",
"providerId": "registration-recaptcha-action",

"""



import unittest, time
from kcapi import OpenID, Keycloak
from .testbed import TestBed, KcBaseTestCase

ADMIN_USER = "admin"
ADMIN_PSW  = "admin1234"
REALM = "test_heroes_test"
ENDPOINT = 'https://sso-cvaldezr-stage.apps.sandbox-m2.ll9k.p1.openshiftapps.com'


TEST_REALM = "TESTING"

class Testing_Mappings(KcBaseTestCase):
    def testing_adding_removing_roles_to_group(self):
        groups = self.kc.build('groups', self.realm)
        self.assertTrue(hasattr(groups, "realmRoles"))

        #TestBed class will create one group called "DC"
        #And three roles called [level-1, level-2, level-3]
        # Initially group has 0 role assigned
        group = {"key":"name", "value": self.groupName}
        roles_mapping = groups.realmRoles(group)
        rolesInGroup = roles_mapping.all()
        self.assertEqual(len(rolesInGroup), 0, 'should be 0 roles initially')

        # assign roles to group
        state = roles_mapping.add(self.roleNames)
        self.assertTrue(state)

        rolesInGroup = roles_mapping.all()
        self.assertEqual(len(rolesInGroup), 3, 'should be 3 roles: [level-1, level-2, level-3]')

        # remove_roles_from_group
        roles_mapping.remove(["level-1", "level-2"])
        rolesInGroup = roles_mapping.all()
        self.assertEqual(len(rolesInGroup), 1, 'we should got only 1 role: [level-3]')

    def testing_adding_removing_user_to_group(self):
        users = self.kc.build('users', self.realm)
        self.assertTrue(hasattr(users, "joinGroup"))

        usr = {"key": "username", "value": "batman"}
        gpr = {"key": "name", "value": self.groupName}

        join_state = users.joinGroup(usr, gpr).isOk()
        self.assertTrue(join_state)

        # def testing_user_group_mapping(self):
        #users = self.kc.build('users', self.realm)
        #usr = {"key": "username", "value": "batman"}
        batman = users.groupMapping(usr)
        self.assertEqual(len(batman.all()), 1, 'the user batman should belong to one group')

        # def testing_user_leaving_group(self):
        # users = self.kc.build('users', self.realm)
        self.assertTrue(hasattr(users, "joinGroup"))

        leave_status = users.leaveGroup(usr, gpr).isOk()
        self.assertTrue(leave_status)

        groups = users.groups(usr)
        self.assertEqual(len(groups), 0)

    def setUp(self):
        super().setUp(create_all=False)
        # self.testbed = TestBed(REALM, ADMIN_USER, ADMIN_PSW, ENDPOINT)
        self.testbed.createRealms()
        self.testbed.createUsers()
        self.testbed.createGroups()
        self.kc = self.testbed.getKeycloak()
        self.realm = self.testbed.REALM
        self.master_realm = self.testbed.getAdminRealm()
        self.roleNames = self.testbed.roleNames
        self.groupName = self.testbed.groupName

    def tearDown(self):
        if self.master_realm.exist(TEST_REALM):
            self.master_realm.remove(TEST_REALM)
        super().tearDown()


if __name__ == '__main__':
    unittest.main()



import json
from .resp import ResponseHandler
from .kcsession import KcSession


class KeycloakCRUD(object):
    @classmethod
    def get_child(cls, that, resource_id, resource_name):
        # .get_child in derived classes need to return instance of derived class
        kc = cls()
        kc.token = that.token
        kc.targets = that.targets.copy()

        kc.targets.addResources([resource_id, resource_name])

        return kc

    def __init__(self): 
        self.targets = None
        self.token = None

    def headers(self):

        if self.token.expired():
            self.token = self.token.refresh()

        return {
                'Content-type': 'application/json', 
                'Authorization': 'Bearer '+ self.token.get_token()
        }

    def setIdentifier(self, _id = None, url = None):
        if _id:
            return url.addResource(_id)
        else:
            return url
    
    def create(self, payload):
        url = self.targets.url('create')

        with KcSession() as session:
            ret = session.post(url, data=json.dumps(payload), headers=self.headers())
        return ResponseHandler(url, method='Post', payload=payload).handleResponse(ret)

    def update(self, obj_id=None, payload=None):
        url = self.targets.url('update')
        target = str(self.setIdentifier(obj_id, url))

        with KcSession() as session:
            ret = session.put(target, data=json.dumps(payload), headers=self.headers())
        return ResponseHandler(target, method='Put', payload=payload).handleResponse(ret)

    def remove(self, _id, payload=None):
        delete = self.targets.url('delete')
        url = self.setIdentifier(_id, delete)
        with KcSession() as session:
            ret = session.delete(url, data=json.dumps(payload), headers=self.headers())
        return ResponseHandler(url, method='Delete', payload=payload).handleResponse(ret)
        
    def get(self, _id):
        url = self.targets.url('read')
        with KcSession() as session:
            ret = session.get(str(self.setIdentifier(_id, url)), headers=self.headers())
        return ResponseHandler(url, method='Get').handleResponse(ret)

    def findAll(self, params=None):
        url = self.targets.url('read')
        with KcSession() as session:
            ret = session.get(url, params=params, headers=self.headers())
        return ResponseHandler(url, method='Get').handleResponse(ret)

    def findFirst(self, params):
        return self.findFirstByKV(params['key'], params['value'])

    def findFirstByKV(self, key, value):
        rows = self.findAll().verify().resp().json()

        for row in rows:
            # Some components do not have Name attribute.
            if key in row and row[key].lower() == value.lower():
                return row

        return []

    def all(self, params=None):
        return self.findAll(params).verify().resp().json()

    def updateUsingKV(self, key, value, obj): 
        res_data = self.findFirstByKV(key,value)

        if res_data: 
            data_id = res_data['id']
            res_data.update(obj)
            return self.update(data_id, res_data).isOk() 
        else:
            return False

    def removeFirstByKV(self, key, value, custom_key="id"):
        row = self.findFirstByKV(key,value)

        if row:
            return self.remove(row[custom_key]).isOk()
        else:
            return False

    def existByKV(self, key, value): 
        ret = self.findFirstByKV(key, value)
        return ret != []

    def exist(self, _id):
        return self.get(_id).ok()

    def get_one(self, obj_id):
        """
        Returns a dict with response content.
        """
        return self.get(obj_id).verify().resp().json()

    def update_rmw(self, obj_id=None, partial_payload=None):
        """
        read/modify/write update - existing object will be partially updated.
        We first read current state, then send to server new state.

        Calling .update() on realm object with partial data can destroy some realm attributes.
        See https://github.com/justinc1/keycloak-fetch-bot/commit/c68610c671193aaf2fe71c23b68e3af3c989db2f.
        """
        if not partial_payload:
            partial_payload = {}
        full_payload = self.get_one(obj_id)
        full_payload.update(partial_payload)
        return self.update(obj_id, full_payload)
